#include "types.H"
#include "kalman.H"
#include "kalmanPVSM.H"
#include "kalmanPVASM.H"
#include "kalmanPVMMV.H"
#include "kalmanPVAMMV.H"
#include "kalmanPVAMMA.H"
#include "kalmanPVAMMVA.H"

KalmanPredictor::KalmanPredictor()
{
  // create all of our different filter-predictors
  _pvsm    =  new KalmanPVSM(); 
  _pvasm   =  new KalmanPVASM();
  _pvmmv   =  new KalmanPVMMV();
  _pvammv  =  new KalmanPVAMMV();
  _pvamma  =  new KalmanPVAMMA();
  _pvammva =  new KalmanPVAMMVA();
}


// destructor
KalmanPredictor::~KalmanPredictor()
{
  delete _pvsm;
  delete _pvasm;
  delete _pvmmv;
  delete _pvammv;
  delete _pvamma;
  delete _pvammva;
}

void
KalmanPredictor::prepare()
{
  // get some things we need from the bundle and send them to
  // the appropriate filter
  double spd = (double)_paramb->getSpecD();
  double delta = 1.0 / _paramb->getCurSRate();
  double mnoisep = _paramb->getPosVar();
  double mnoiseq = _paramb->getQuatVar();
  double vdelta = 1.0 / _paramb->getOrigSRate();
  double ukfflag = _paramb->getUKFFlag();

  switch (_paramb->getKType()) {
    case tPVSM:
      _pvsm->initialize(spd, delta, mnoisep, mnoiseq, ukfflag);
      break;
    case tPVASM:
      _pvasm->initialize(spd, delta, mnoisep, mnoiseq, ukfflag);
      break;     
    case tPVMMV:
      _pvmmv->initialize(spd, delta, mnoisep, mnoiseq, delta, ukfflag);
      break; 
    case tPVAMMV:
      _pvammv->initialize(spd, delta, mnoisep, mnoiseq, delta, ukfflag);
      break; 
    case tPVAMMA:
      _pvamma->initialize(spd, delta, mnoisep, mnoiseq, delta, ukfflag);
      break; 
    case tPVAMMVA:
      _pvammva->initialize(spd, delta, mnoisep, mnoiseq, delta, ukfflag);
      break; 
  default: 
    cerr << "No Kalman Predictor Selected!!" << endl;
    
  }
  
}

Pose *
KalmanPredictor::predict(Pose *userpos, double pinterval)
{
   switch (_paramb->getKType()) {
    case tPVSM:
      return _pvsm->filter(userpos, pinterval);
      break;
   case tPVASM:
      return _pvasm->filter(userpos, pinterval);
      break;
   case tPVMMV: 
      return _pvmmv->filter(userpos, pinterval);
      break;
   case tPVAMMV:
      return _pvammv->filter(userpos, pinterval);
      break;
   case tPVAMMA:
      return _pvamma->filter(userpos, pinterval);
      break;
   case tPVAMMVA:
      return _pvammva->filter(userpos, pinterval);
      break;
    default:
      cerr << "No Predictor Selected!!!" << endl;
      // for testing, just make everythin constant
      Pose *tmp = new Pose(userpos->_tstamp+pinterval, Vec3d(0.0, 0.0, 0.0), Quat(0.0,0.0,0.0,1.0));
      return tmp;
   }
}



