// Code for the local polynomial predictor

#include "lpoly.H"
#include "types.H"
#include "mmatrix.H"
#include "vvector.H"

#define SPEC_DENSITY 1

// Setup storage for simple PVSM KF and EKF
LPoly::LPoly() : KalmanBase(), ExtendedKalmanBase(), _numwin(0),
                 _degree(0), _posewin(50), _ctr(0)
{
  // create for KF
  _RR = MMatrix(1,1);
  _HH = MMatrix(1,2);
  _QQ = MMatrix(2,2);
  _fundMat = MMatrix(2,2);

  for (int i = 0; i < 3; ++i) {
    _PP[i] = MMatrix(2,2); 
    _state[i] = MMatrix(2,1);
  }
  
  // create for EKF
  _ER = MMatrix(4,4);
  _EH = MMatrix(4,7);
  _EQ = MMatrix(7,7);
  _EfundMat = MMatrix(7,7);
  _EP = MMatrix(7,7);
  _Estate = MMatrix(7,1);
  _edt = _spd = 0.0;
}


// destructor
LPoly::~LPoly()
{
  _RR.deleteMatrix();
  _HH.deleteMatrix();
  _QQ.deleteMatrix();
  _fundMat.deleteMatrix();
  for (int i = 0; i < 3; ++i) {
    _PP[i].deleteMatrix(); 
    _state[i].deleteMatrix();
  }

  _ER.deleteMatrix();
  _EH.deleteMatrix();
  _EQ.deleteMatrix();
  _EfundMat.deleteMatrix();
  _EP.deleteMatrix();
  _Estate.deleteMatrix();
 
  for (int i = 0; i < _posewin.num(); ++i)
     delete _posewin[i];
}

void
LPoly::prepare()
{
  // collect what we need from the bundle
  _numwin = _paramb->getNumPts();
  _degree = _paramb->getDegree();
  _ctr = 0;

  // if the Kalman flag is set we will Kalman filter
  if (_paramb->getKFFlag()) {
     double dt = 1.0 / _paramb->getCurSRate();
     double np = _paramb->getPosVar();
     double nq = _paramb->getQuatVar();
     double spd = SPEC_DENSITY;
  
     // initialize for KF
     _HH(0,0) = 1.0;
     _RR(0,0) = np;

     _fundMat(0,0) = _fundMat(1,1) = 1.0;
     _fundMat(0,1) = dt; 


     _QQ(0,0) = ((dt*dt*dt) / 3.0) * spd;
     _QQ(0,1) = _QQ(1,0) = ((dt*dt) / 2.0) * spd;
     _QQ(1,1) = dt * spd;

     for (int i = 0; i < 3; ++i) {
	_PP[i].setZeros();
	_PP[i](0,0) = _PP[i](1,1) = INITVAR;
	// zero out the state vectors
	_state[i](0,0) = _state[i](1,0) = 0.0;
     }
     
     // initialize EKF
     _edt = dt;
     _spd = spd;
     // we can set the measurment noise covariance matrix here (at least that can
     // stay constant	
     _ER.setZeros();
   
     for (int i = 0; i < 4; ++i)
      _ER(i,i) = nq;
     _EP.setZeros();
     for (int i = 0; i < 7; ++i) {
	if (i < 4)
	   _EP(i,i) = 1.0;
	else
	   _EP(i,i) = INITVAR;
	_Estate(i,0) = 0.0; 
     }  

  }

  // reset the window list
   for (int i = 0; i < _posewin.num(); ++i)
     delete _posewin[i];
  _posewin.clear();
}


Pose *
LPoly::predict(Pose *userpos, double pinterval)
{
   Pose *npos = 0;

   if (_paramb->getKFFlag()) {
     if (_ctr <= MAXBEGIN-6) 
       ++_ctr;
     if (_ctr == 1) {
	for (int i = 0; i < 3; ++i) 
	   (_state[i])(0,0) = userpos->_pos[i];
	for (int i = 0; i < 3; ++i)
	   _Estate(i,0) = userpos->_orient.V()[i];
	_Estate(3,0) = userpos->_orient.S();
        return userpos;
     }
      // Kalman Filter
      kalmanFilter(userpos, 1, 2, kPVSM);
      extendedKalmanFilter(userpos, 4, 7, kPVSM);
      if (_ctr < MAXBEGIN-6)
        return userpos; 
      else { // now start filling the pose records for prediction  
         Vec3d nv((_state[0])(0,0), (_state[1])(0,0), (_state[2])(0,0));
         Quat nq(_Estate(0,0), _Estate(1,0), _Estate(2,0), _Estate(3,0)); 
	 npos = new Pose(userpos->_tstamp, nv, normalize(nq));
      }
   }
   
   Pose *tmppos;
   if (_paramb->getKFFlag())
     tmppos = new Pose(npos);
   else
     tmppos = new Pose(userpos);

   // create and update the pose window
      if (_posewin.num() < _numwin) {
	 _posewin += tmppos;
	 return userpos;
      }
      else {
         Pose *p;
	 for (int i = 1; i < _numwin; ++i) {
            if (i == 1)
	       p = _posewin[0];  
	    _posewin[i-1] = _posewin[i];            
         }
         delete p;
	 _posewin.pop();
	 _posewin += tmppos;
      }

    // predict
    switch(_paramb->getLPolyType()) {
      case tlagrangian: {    
                           return lagrangian(pinterval);
      }
      case tlsquares:   {
                           return leastSquares(pinterval);
      }
      case tcspline:    { 
                           return cubicSpline(pinterval);
      }
      case ttaylor:     {
                           return taylorExp(pinterval);
      }
      default: {
        cerr << "No Predictor Selected!!!" << endl;
        // for testing, just make everythin constant
        Pose *tmp = new Pose(userpos->_tstamp+pinterval, Vec3d(0.0, 0.0, 0.0), Quat(0.0,0.0,0.0,1.0));
        return tmp;
      }
    }
}

// various polynomial predictors

Pose *
LPoly::lagrangian(double pinterval)
{
 double lj = 0.0;
 Vec3d posp(0.0,0.0,0.0);
 Quat  rotp(0.0,0.0,0.0,0.0);

 for(int j = 0; j < _posewin.num(); ++j) {
   // initialize the lagrangain polys
   lj = 1.0;
   for (int k = 0; k < _posewin.num(); ++k) {
     if (k != j) // compute lagrange poly iteratively
      lj *= (((_posewin[_numwin-1]->_tstamp + pinterval) - _posewin[k]->_tstamp) /
             (_posewin[j]->_tstamp - _posewin[k]->_tstamp));
   }
   posp += (lj * _posewin[j]->_pos); // predicted position
   rotp += (lj * _posewin[j]->_orient); // predicted orientation
 }  

 // make sure to renormalize the quaternion
 return new Pose(_posewin[_numwin-1]->_tstamp + pinterval, posp, normalize(rotp));
}


Pose *
LPoly::leastSquares(double pinterval)
{
  // error check
  if ((_degree+1 >  _numwin) || (_numwin != _posewin.num())) {
    cerr << "Problem, degree higher than num points!!!" << endl;
    return new Pose();
  }
 
  // create data structures we need
  double *q     = new double[_degree+1]; 
  double *alpha = new double[_degree+1];
  double *beta  = new double[_degree+1];
  double *gamma = new double[_degree+1];
  MMatrix *Qi   = new MMatrix(_degree+1, _numwin); // need to store intermediate q's
  Vec3d  *ap    = new Vec3d[_degree+1];
  Quat   *aq    = new Quat[_degree+1];

  // base case
  q[0] = 1.0;
  int i;
  for(i = 0; i < _numwin; ++i) 
    (*Qi)(0,i) = 1.0;
  
  double ttot = 0.0;
  for(i = 0; i < _numwin; ++i)
   ttot += _posewin[i]->_tstamp;

  gamma[0] = _numwin;
  
  if (_degree >= 1) {
    q[1] = (pinterval+_posewin[_numwin-1]->_tstamp) - (1.0 / _numwin)*ttot;
    // store q values for ti's
    for(i = 0; i < _numwin; ++i)
      (*Qi)(1,i) = _posewin[i]->_tstamp - (1.0 / _numwin)*ttot;
    gamma[1] = 0.0;  
    for (i = 0; i < _numwin; ++i)
      gamma[1] += ((*Qi)(1,i) * (*Qi)(1,i));
    alpha[1] = (1.0 / _numwin)*ttot;
  }
   
  // iteratively solve the recurrence relation
  int j;
  for (j = 1; j <= _degree-1; ++j) {
    // compute the jth gamma
    double tmpgamma = 0.0;
    for(i = 0; i < _numwin; ++i)
      tmpgamma += ((*Qi)(j,i)*(*Qi)(j,i));
    gamma[j] = tmpgamma;
    
    // compute the j+1th alpha
    double tmpalpha = 0.0;
    for(i = 0; i < _numwin; ++i)
      tmpalpha += _posewin[i]->_tstamp * ((*Qi)(j,i)*(*Qi)(j,i));
    alpha[j+1] = (1.0 / gamma[j]) * tmpalpha;
    
    // compute jth beta
    beta[j] = gamma[j] / gamma[j-1];

    // compute new j+1th q for the predicted time
    q[j+1] = (_posewin[_numwin-1]->_tstamp + pinterval) * q[j] - (alpha[j+1] * q[j]) -
             (beta[j]*q[j-1]);

    // compute the j+1th q's for the ti's and store in Qi
    for (i = 0; i < _numwin; ++i)
      (*Qi)(j+1,i) = (_posewin[i]->_tstamp * (*Qi)(j,i)) - (alpha[j+1] * (*Qi)(j,i)) -
                     (beta[j] * (*Qi)(j-1,i));
  }

  // compute the coefficients (ak's)
  int k;
  for(k = 0; k <= _degree; ++k) {
    Vec3d tmppos(0.0,0.0,0.0);
    Quat  tmpqrot(0.0,0.0,0.0,0.0);
    double tmp = 0.0;
    for (i = 0; i < _numwin; ++i) {
      tmppos += (*Qi)(k,i) * _posewin[i]->_pos;
      tmpqrot += (*Qi)(k,i) * _posewin[i]->_orient;
      tmp += ((*Qi)(k,i) * (*Qi)(k,i));
    }
    gamma[k] = tmp;
    ap[k] = (1.0 / gamma[k]) * tmppos;
    aq[k] = (1.0 / gamma[k]) * tmpqrot;
  }

  Vec3d ppos(0.0,0.0,0.0);
  Quat  prot(0.0,0.0,0.0,0.0);
  // finally compute predicted pose
  for (k = 0; k <= _degree; ++k) {
   ppos += q[k] * ap[k];
   prot += q[k] * aq[k];
  }

  // delete memory
  delete [] q;
  delete [] alpha;
  delete [] beta;
  delete [] gamma;
  delete [] ap;
  delete [] aq;
  Qi->deleteMatrix();
  delete Qi;

  // make sure to renormalize the quaternion
 return new Pose(_posewin[_numwin-1]->_tstamp + pinterval, ppos, normalize(prot));

}


Pose *
LPoly::cubicSpline(double pinterval)
{
  // quick error check
  if (_numwin < 2) {
    cerr << "Problem with window size" << endl;
    return new Pose();
  }

  Vec3d posacclo(0.0,0.0,0.0);
  Quat  angacclo(0.0,0.0,0.0,1.0);  
  Vec3d posacchi(0.0,0.0,0.0);
  Quat  angacchi(0.0,0.0,0.0,1.0);  
  Vec3d poslo(0.0,0.0,0.0);
  Vec3d poshi(0.0,0.0,0.0);
  Quat  orientlo(0.0,0.0,0.0,1.0);
  Quat  orienthi(0.0,0.0,0.0,1.0);
  double tstmplo = 0.0;
  double tstmphi = 0.0;
  if (_paramb->getDerFly()) {
    // we know a few things already
    poslo = _posewin[4]->_pos;
    poshi = _posewin[_numwin-1]->_pos;  
    orientlo = _posewin[4]->_orient;
    orienthi = _posewin[_numwin-1]->_orient;
    tstmplo = _posewin[4]->_tstamp;
    tstmphi = _posewin[_numwin-1]->_tstamp;

    // now compute second derivatives for position and orientation
    
    Vec3d tmp1 = 3.0 * _posewin[4]->_pos;
    Vec3d tmp2 = -9.0 * _posewin[3]->_pos;
    Vec3d tmp3 = 10.0 * _posewin[2]->_pos;
    Vec3d tmp4 = -5.0 * _posewin[1]->_pos;
    Vec3d tmp5 = _posewin[0]->_pos;

    double srate = _paramb->getCurSRate();

    posacclo = (11.0 / (12.0 * (1.0 / (srate * srate)))) *
             (tmp1 + tmp2 + tmp3 + tmp4 + tmp5);   

    tmp1 = 3.0 * _posewin[_numwin-1]->_pos;
    tmp2 = -9.0 * _posewin[_numwin-2]->_pos;
    tmp3 = 10.0 * _posewin[_numwin-3]->_pos;
    tmp4 = -5.0 * _posewin[_numwin-4]->_pos;
    tmp5 = _posewin[_numwin-5]->_pos;
    
    posacchi = (11.0 / (12.0 * (1.0 / (srate * srate)))) *
             (tmp1 + tmp2 + tmp3 + tmp4 + tmp5);   
    //  
    Quat qtmp1 = 3.0 * _posewin[4]->_orient;
    Quat qtmp2 = -9.0 * _posewin[3]->_orient;
    Quat qtmp3 = 10.0 * _posewin[2]->_orient;
    Quat qtmp4 = -5.0 * _posewin[1]->_orient;
    Quat qtmp5 = _posewin[0]->_orient;

    angacclo = (11.0 / (12.0 * (1.0 / (srate * srate)))) *
             (qtmp1 + qtmp2 + qtmp3 + qtmp4 + qtmp5);   

    qtmp1 = 3.0 * _posewin[_numwin-1]->_orient;
    qtmp2 = -9.0 * _posewin[_numwin-2]->_orient;
    qtmp3 = 10.0 * _posewin[_numwin-3]->_orient;
    qtmp4 = -5.0 * _posewin[_numwin-4]->_orient;
    qtmp5 = _posewin[_numwin-5]->_orient;
    
    angacchi = (11.0 / (12.0 * (1.0 / (srate * srate)))) *
               (qtmp1 + qtmp2 + qtmp3 + qtmp4 + qtmp5);   

  }
  else {
   poslo = _posewin[0]->_pos;
   poshi = _posewin[_numwin-1]->_pos;
   orientlo = _posewin[0]->_orient;
   orienthi = _posewin[_numwin-1]->_orient;
   posacclo = _posewin[0]->_posacc;
   posacchi = _posewin[_numwin-1]->_posacc;
   
   angacclo = (0.5 * _posewin[0]->_orient *_posewin[0]->_angacc) + 
              (0.25 * _posewin[0]->_orient * _posewin[0]->_angvel * 
              _posewin[0]->_angvel); 

   angacchi = (0.5 * _posewin[_numwin-1]->_orient *_posewin[_numwin-1]->_angacc) + 
              (0.25 * _posewin[_numwin-1]->_orient * _posewin[_numwin-1]->_angvel * 
              _posewin[_numwin-1]->_angvel);

   tstmplo = _posewin[0]->_tstamp;
   tstmphi = _posewin[_numwin-1]->_tstamp;
  }

  // so the first and last records in the posewin will be our high an lows
  double h = tstmphi - tstmplo;

  double a = -pinterval / h;
  double b = ((tstmphi + pinterval) - tstmplo) / h;

  Vec3d ppos = (a * poslo) +
               (b * poshi) +
               ((a*a*a-a)*posacclo +
                (b*b*b-b)*posacchi) *
                (h*h) / 6.0;

   Quat prot = (a * orientlo) +
               (b * orienthi) +
               ((a*a*a-a)*angacclo  +
                (b*b*b-b)*angacchi) *
                (h*h) * (1.0 / 6.0);

  

  return new Pose(_posewin[_numwin-1]->_tstamp + pinterval, ppos, normalize(prot));
}


Pose *
LPoly::taylorExp(double pinterval)
{
  Vec3d posvel(0.0,0.0,0.0);
  Vec3d posacc(0.0,0.0,0.0);
  Quat angvel(0.0,0.0,0.0,1.0);
  Quat angacc(0.0,0.0,0.0,1.0);

  if (_paramb->getDerFly()) {
    // first compute 1st and 2nd derivative for the position vector
    // using 4th order explicit backward differencing
    Vec3d tmp1 = 25.0 * _posewin[_numwin-1]->_pos;
    Vec3d tmp2 = -48.0 * _posewin[_numwin-2]->_pos;
    Vec3d tmp3 = 36.0 * _posewin[_numwin-3]->_pos;
    Vec3d tmp4 = -16.0 * _posewin[_numwin-4]->_pos;
    Vec3d tmp5 = 3.0 * _posewin[_numwin-5]->_pos;
 
    double srate = _paramb->getCurSRate();
    posvel = (1.0 / (12.0 * (1.0 / srate))) *
             (tmp1 + tmp2 + tmp3 + tmp4 + tmp5);   
  
    tmp1 = 3.0 * _posewin[_numwin-1]->_pos;
    tmp2 = -9.0 * _posewin[_numwin-2]->_pos;
    tmp3 = 10.0 * _posewin[_numwin-3]->_pos;
    tmp4 = -5.0 * _posewin[_numwin-4]->_pos;
    tmp5 = _posewin[_numwin-5]->_pos;

    posacc = (11.0 / (12.0 * (1.0 / (srate * srate)))) *
             (tmp1 + tmp2 + tmp3 + tmp4 + tmp5);   

    // now for first and second derivatives for quaternions
    Quat qtmp1 = 25.0 * _posewin[_numwin-1]->_orient;
    Quat qtmp2 = -48.0 * _posewin[_numwin-2]->_orient;
    Quat qtmp3 = 36.0 * _posewin[_numwin-3]->_orient;
    Quat qtmp4 = -16.0 * _posewin[_numwin-4]->_orient;
    Quat qtmp5 = 3.0 * _posewin[_numwin-5]->_orient;

    Quat qtmp6 = (1.0 / (12.0 * (1.0 / srate))) *
             (qtmp1 + qtmp2 + qtmp3 + qtmp4 + qtmp5);   
    
    angvel = 2.0 * inverse(_posewin[_numwin-1]->_orient) *qtmp6;

    qtmp1 = 25.0 * inverse(_posewin[_numwin-1]->_orient);
    qtmp2 = -48.0 * inverse(_posewin[_numwin-2]->_orient);
    qtmp3 = 36.0 * inverse(_posewin[_numwin-3]->_orient);
    qtmp4 = -16.0 * inverse(_posewin[_numwin-4]->_orient);
    qtmp5 = 3.0 * inverse(_posewin[_numwin-5]->_orient);

    Quat qtmp7 = (1.0 / (12.0 * (1.0 / srate))) *
             (qtmp1 + qtmp2 + qtmp3 + qtmp4 + qtmp5);   
    Quat qparta = 2.0 * qtmp7 * qtmp6; //_posewin[_numwin-1]->_orient;
  
    qtmp1 = 3.0 * _posewin[_numwin-1]->_orient;
    qtmp2 = -9.0 * _posewin[_numwin-2]->_orient;
    qtmp3 = 10.0 * _posewin[_numwin-3]->_orient;
    qtmp4 = -5.0 * _posewin[_numwin-4]->_orient;
    qtmp5 = _posewin[_numwin-5]->_orient;

    qtmp6 = (11.0 / (12.0 * (1.0 / (srate * srate)))) *
             (qtmp1 + qtmp2 + qtmp3 + qtmp4 + qtmp5);   
 
    Quat qpartb = 2.0 * inverse(_posewin[_numwin-1]->_orient) * qtmp6;

    angacc = (qparta + qpartb);

  }
  else {
    // take derivatives from initial calculations   
    posvel = _posewin[_numwin-1]->_posvel;
    posacc = _posewin[_numwin-1]->_posacc;
    angvel = _posewin[_numwin-1]->_angvel;
    angacc = _posewin[_numwin-1]->_angacc;
  }

  // first we do the taylor expansion for position
    Vec3d ppos(0.0,0.0,0.0);
    Quat  prot(0.0,0.0,0.0,1.0);
    
    ppos = _posewin[_numwin-1]->_pos +
      (pinterval * posvel) +
      (pinterval * pinterval * 0.5 * posacc);
    
    // send we do rotation 
    Quat a = _posewin[_numwin-1]->_orient;
    Quat b = (pinterval * 0.5) * a * angvel;
    Quat c = ((-1.0 / 8.0) * pinterval * pinterval * 
              angvel.length()  *
              angvel.length()) * a;
    Quat d = (0.25 * pinterval * pinterval) * a * angacc;

    prot = a + b + c + d;

    // make sure to renormalize the quaternion
    return new Pose(_posewin[_numwin-1]->_tstamp + pinterval, ppos, normalize(prot));
}
 

