// functions will perform KF and EKF filtering used in MMAE 
// also create and intialize data

#include "mmaebase.H"


MMAEKalmanBase::MMAEKalmanBase(KALMANMODEL mod) : _mod(mod)
{
   // initialize variables according to kalman model
   switch(mod) {
    case kPSM: 
      _HH       = MMatrix(3,3);
      _QQ       = MMatrix(3,3);
      _RR       = MMatrix(3,3);
      _fundMat  = MMatrix(3,3);
      _PP       = MMatrix(3,3);
      _state    = MMatrix(3,1);
      _AA       = MMatrix(3,3);
      _residual = MMatrix(3,1);
      break;
    case kPVSM:
      _HH       = MMatrix(3,6);
      _QQ       = MMatrix(6,6);
      _RR       = MMatrix(3,3);
      _fundMat  = MMatrix(6,6);
      _PP       = MMatrix(6,6);
      _state    = MMatrix(6,1);
      _AA       = MMatrix(3,3);
      _residual = MMatrix(3,1);
      break;
    case kPVASM:     
      _HH       = MMatrix(3,9);
      _QQ       = MMatrix(9,9);
      _RR       = MMatrix(3,3);
      _fundMat  = MMatrix(9,9);
      _PP       = MMatrix(9,9);
      _state    = MMatrix(9,1);
      _AA       = MMatrix(3,3);
      _residual = MMatrix(3,1); 
      break;
    default: 
      cerr << "Problem in MMAEKalmanBaseCompWise constructor" << endl;
   }

}


// destructor
MMAEKalmanBase::~MMAEKalmanBase()
{
   _HH.deleteMatrix();
   _QQ.deleteMatrix();
   _RR.deleteMatrix();
   _fundMat.deleteMatrix();
   _PP.deleteMatrix();
   _state.deleteMatrix();
   _AA.deleteMatrix();
   _residual.deleteMatrix();

}


void 
MMAEKalmanBase::initialize(double spd, double dt, double noisep)
{
   switch(_mod) {
    case kPSM:
      _HH.setIdentity();
      _fundMat.setIdentity();
      _QQ.setIdentity();
      _QQ.multScalar(spd*dt);
      _RR.setIdentity();
      _RR.multScalar(noisep);  
      _PP.setZeros();
   
      _AA.setZeros();
      _residual.setZeros();
      for (int i = 0; i < 3; ++i) {
	 _PP(i,i) = INITVAR;
	 _state(i,0) = 0.0;
      }
      break;
    case kPVSM:
      _HH(0,0) = _HH(1,2) = _HH(2,4) = 1.0;
      _RR(0,0) = _RR(1,1) = _RR(2,2) = noisep;
 
      _fundMat.setIdentity();
      _fundMat(0,1) = _fundMat(2,3) = _fundMat(4,5) = dt;

      _QQ(0,0) = _QQ(2,2) = _QQ(4,4) = ((dt*dt*dt) / 3.0) * spd;
      _QQ(0,1) = _QQ(1,0) = ((dt*dt) / 2.0) * spd;
      _QQ(2,3) = _QQ(3,2) = ((dt*dt) / 2.0) * spd;
      _QQ(4,5) = _QQ(5,4) = ((dt*dt) / 2.0) * spd;
      _QQ(1,1) = _QQ(3,3) = _QQ(5,5) = dt * spd;

      _PP.setZeros();
      _AA.setZeros();
      _residual.setZeros();
      for (int i = 0; i < 6; ++i) {
	 _PP(i,i) = INITVAR;
	 _state(i,0) = 0.0;
      }
      break;
    case kPVASM:
      _HH(0,0) = _HH(1,3) = _HH(2,5) = 1.0;
      _RR(0,0) = _RR(1,1) = _RR(2,2) = noisep;

      _fundMat.setIdentity();
      _fundMat(0,1) = _fundMat(1,2) = _fundMat(3,4) = 
      _fundMat(4,5) = _fundMat(6,7) = _fundMat(7,8) = dt;
      _fundMat(0,2) = _fundMat(3,5) = _fundMat(6,8) = dt*dt/2.0;
      
      _QQ(0,0) = _QQ(3,3) = _QQ(6,6) = (pow(dt,5) / 20.0) * spd;
      _QQ(1,1) = _QQ(4,4) = _QQ(7,7) = (pow(dt,3) / 3.0) * spd;
      _QQ(2,2) = _QQ(5,5) = _QQ(8,8) = dt * spd;
      _QQ(0,1) = _QQ(1,0) = (pow(dt,4) / 8.0) * spd;
      _QQ(0,2) = _QQ(2,0) = (pow(dt,3) / 6.0) * spd;
      _QQ(1,2) = _QQ(2,1) = dt*dt*0.5 * spd;   
      _QQ(3,4) = _QQ(4,3) = (pow(dt,4) / 8.0) * spd;
      _QQ(3,5) = _QQ(5,3) = (pow(dt,3) / 6.0) * spd;
      _QQ(4,5) = _QQ(5,4) = dt*dt*0.5 * spd;   
      _QQ(6,7) = _QQ(7,6) = (pow(dt,4) / 8.0) * spd;
      _QQ(6,8) = _QQ(8,6) = (pow(dt,3) / 6.0) * spd;
      _QQ(7,8) = _QQ(8,7) = dt*dt*0.5 * spd;   

      _PP.setZeros();
      _AA.setZeros();
      _residual.setZeros();
      for (int i = 0; i < 9; ++i) {
	 _PP(i,i) = INITVAR;
	 _state(i,0) = 0.0;
      }
      break;
    default: 
      cerr << "Problem in MMAEKalmanBase Initialize" << endl;
   }

}


void MMAEKalmanBase::resetPP()
{
   _PP.setZeros();
   switch(_mod) {
     case kPSM: 
       for (int i = 0; i < 3; ++i) 
	  _PP(i,i) = INITVAR; 
       break;
     case kPVSM:
        for (int i = 0; i < 6; ++i) 
	  _PP(i,i) = INITVAR; 
        break; 
     case kPVASM:       
        for (int i = 0; i < 9; ++i) 
	  _PP(i,i) = INITVAR; 
        break; 
     default:
       cerr << "Problem in ResetPP" << endl;
   }

}



// do the filtering
void MMAEKalmanBase::mmaeKalmanFilter(Pose *userpos, int zNum)
{
   // do the time update step
   MMatrix pstate = _fundMat * _state;

   // do the a priori extimate of the the process covariance matrix
   MMatrix pcov = _fundMat * (_PP * transp(_fundMat)) + _QQ;

   // compute the Kalman Gain
   _AA = _HH * (pcov * transp(_HH)) + _RR;

   MMatrix KG = pcov * (transp(_HH) * minverse2(_AA));

   // do the measurment update step
   MMatrix Z(zNum,1);
   switch(_mod) {
    case kPSM:
    case kPVSM:
    case kPVASM: 
      Z(0,0) = userpos->_pos[0];
      Z(1,0) = userpos->_pos[1];
      Z(2,0) = userpos->_pos[2];
      break;
    default: ;
   }

   _residual = Z - (_HH * pstate);
   _state = pstate + KG * (_residual);

   // update the covariance matrix
   _PP = pcov - (KG * _HH * pcov);

   //    cerr << " " << _state(0,0) << " " << sqrt(_PP(0,0)) << endl;
}


MMAEKalmanBaseCompWise::MMAEKalmanBaseCompWise(KALMANMODEL mod) : _mod(mod)
{
   // initialize variables according to kalman model
   switch(mod) {
    case kPSM: 
      _HH       = MMatrix(1,1);
      _QQ       = MMatrix(1,1);
      _RR       = MMatrix(1,1);
      _fundMat  = MMatrix(1,1);
      for (int i = 0; i < 3; ++i) {
	 _PP[i]       = MMatrix(1,1);
	 _state[i]    = MMatrix(1,1);
	 _AA[i]       = MMatrix(1,1);
	 _residual[i] = MMatrix(1,1);
      }
      break;
    case kPVSM:
      _RR = MMatrix(1,1);
      _HH = MMatrix(1,2);
      _QQ = MMatrix(2,2);
      _fundMat = MMatrix(2,2);

      for (int i = 0; i < 3; ++i) {
	 _PP[i]       = MMatrix(2,2); 
	 _state[i]    = MMatrix(2,1);
	 _AA[i]       = MMatrix(1,1);
	 _residual[i] = MMatrix(1,1);
      }
      break;
    case kPVASM:     
      _RR = MMatrix(1,1);
      _QQ = MMatrix(3,3);
      _HH = MMatrix(1,3);
      _fundMat = MMatrix(3,3);

      for (int i = 0; i < 3; ++i) {
        _PP[i]       = MMatrix(3,3);
        _state[i]    = MMatrix(3,1);
        _AA[i]       = MMatrix(1,1);
	_residual[i] = MMatrix(1,1);
      }
      break;
    default: 
      cerr << "Problem in MMAEKalmanBaseCompWise constructor" << endl;
   }

}


// destructor
MMAEKalmanBaseCompWise::~MMAEKalmanBaseCompWise()
{
  _RR.deleteMatrix();
  _QQ.deleteMatrix();
  _HH.deleteMatrix();      
  _fundMat.deleteMatrix();

  for (int i = 0; i < 3; ++i) {
     _PP[i].deleteMatrix();
     _state[i].deleteMatrix();
     _AA[i].deleteMatrix();
     _residual[i].deleteMatrix();
  }
}

void 
MMAEKalmanBaseCompWise::initialize(double spd, double dt, double noisep)
{
   switch(_mod) {
    case kPSM:
      _HH.setIdentity();
      _fundMat.setIdentity();
      _QQ.setIdentity();
      _QQ.multScalar(spd*dt);
      _RR.setIdentity();
      _RR.multScalar(noisep);  
 
      for (int i = 0; i < 3; ++i) {
	 _PP[i].setZeros();
	 _AA[i].setZeros();
	 _residual[i].setZeros(); 
	 _PP[i](0,0) = INITVAR;
	 _state[i](0,0) = 0.0;
      }
      break;
    case kPVSM:
      _HH(0,0) = 1.0;
      _RR(0,0) = noisep;
      _fundMat(0,0) = _fundMat(1,1) = 1.0;
      _fundMat(0,1) = dt; 
      _QQ(0,0) = ((dt*dt*dt) / 3.0) * spd;
      _QQ(0,1) = _QQ(1,0) = ((dt*dt) / 2.0) * spd;
      _QQ(1,1) = dt * spd;

      for (int i = 0; i < 3; ++i) {
	 _PP[i].setZeros();
         _residual[i].setZeros();
         _AA[i].setZeros();
	 _PP[i](0,0) = _PP[i](1,1) = INITVAR;
	 _state[i](0,0) = _state[i](1,0) = 0.0;
      }
      break;
    case kPVASM:
      _HH(0,0) = 1.0;
      _RR(0,0) = noisep;

      _fundMat(0,0) = _fundMat(1,1) = _fundMat(2,2) = 1.0;
      _fundMat(0,1) = _fundMat(1,2) = dt;
      _fundMat(0,2) = dt*dt * 0.5; 
 
      _QQ(0,0) = (pow(dt,5) / 20.0) * spd;
      _QQ(1,1) = (pow(dt,3) / 3.0) * spd;
      _QQ(2,2) = dt * spd;
      _QQ(0,1) = _QQ(1,0) = (pow(dt,4) / 8.0) * spd;
      _QQ(0,2) = _QQ(2,0) = (pow(dt,3) / 6.0) * spd;
      _QQ(1,2) = _QQ(2,1) = dt*dt*0.5 * spd;

      // initialize P and zero out the state v	ector
      for (int i = 0; i < 3; ++i) {
	 _PP[i].setZeros();
         _residual[i].setZeros();
         _AA[i].setZeros();
	 _PP[i](0,0) = _PP[i](1,1) = _PP[i](2,2) = INITVAR;
	 _state[i](0,0) = _state[i](1,0) = _state[i](2,0) = 0.0;
      }
      break;
    default: 
        cerr << "Problem in MMAEKalmanBaseCompWise initialize" << endl;
   }
}



// do the filtering
void MMAEKalmanBaseCompWise::mmaeKalmanFilter(Pose *userpos, int zNum)
{
   for (int i = 0; i < 3; ++i) {
      // do the time update step	
      MMatrix pstate = _fundMat * _state[i];

      // do the a priori extimate of the the process covariance matrix
      MMatrix pcov = _fundMat * (_PP[i] * transp(_fundMat)) + _QQ;
      
      //  compute the Kalman Gain
      _AA[i] = _HH * (pcov * transp(_HH)) + _RR;

      MMatrix KG = pcov * (transp(_HH) * minverse2(_AA[i]));
      
      // do the measurment update step
      MMatrix Z(zNum,1);
      switch(_mod) {
       case kPSM:
       case kPVSM:
       case kPVASM: 
	 Z(0,0) = userpos->_pos[i];
	 break;
       default: ;
      }
      
      _residual[i] = Z - (_HH * pstate);
      _state[i] = pstate + KG * (_residual[i]);
   
      // update the covariance matrix
      _PP[i] = pcov - (KG * _HH * pcov);
      
      // cerr << " " << _state(0,0) << " " << sqrt(_PP(0,0)) << endl;
   }
}


MMAEExtendedKalmanBase::MMAEExtendedKalmanBase(KALMANMODEL mod) : _mod(mod),
                                                                  _spd(0.0),
                                                                  _edt(0.0)
{
// _up = new Updater();

 switch(mod) {
  case kPSM: 
    _Estate = MMatrix(4,1);
    _EP = MMatrix(4,4);
    _EH = MMatrix(4,4);
    _EH2 = MMatrix(4,4);
    _EQ = MMatrix(4,4);
    _ER = MMatrix(4,4);
    _Eresidual = MMatrix(4,1);
    _EA = MMatrix(4,4);
    _EfundMat = MMatrix(4,4);
    break;
  case kPVSM:
    _Estate = MMatrix(7,1);
    _EP = MMatrix(7,7);
    _EH = MMatrix(4,7);
    _EH2 = MMatrix(4,7);
    _EQ = MMatrix(7,7);
    _ER = MMatrix(4,4);
    _Eresidual = MMatrix(4,1);
    _EA = MMatrix(4,4);
    _EfundMat = MMatrix(7,7);
    break;
   case kPVASM:
    _Estate = MMatrix(10,1);
    _EP = MMatrix(10,10);
    _EH = MMatrix(4,10);
    _EH2 = MMatrix(4,10);
    _EQ = MMatrix(10,10);
    _ER = MMatrix(4,4);
    _Eresidual = MMatrix(4,1);
    _EA = MMatrix(4,4);
    _EfundMat = MMatrix(10,10);
    break;
   default: cerr << "Problem with MMAEExtendedKalmanBase Constuctor" << endl;
 }
}


// destructor

MMAEExtendedKalmanBase::~MMAEExtendedKalmanBase()
{
  _Estate.deleteMatrix(); 
  _EP.deleteMatrix();
  _EH.deleteMatrix();
  _EH2.deleteMatrix();
  _EQ.deleteMatrix();
  _ER.deleteMatrix();
  _Eresidual.deleteMatrix();
  _EA.deleteMatrix();
  _EfundMat.deleteMatrix(); 
  
//  if (_up != 0)
  //   delete _up;
}

void 
MMAEExtendedKalmanBase::initialize(double spd, double dt, double noiseq)
{
  _spd = spd;
  _edt = dt;

  switch(_mod) {
   case kPSM:
     _EH2.setZeros();
     _EH2(0,0) = _EH2(1,1) = _EH2(2,2) = _EH2(3,3) = 1.0;
     _ER.setZeros();
     _EA.setZeros();
     _Eresidual.setZeros();
     for (int i = 0; i < 4; ++i)
       _ER(i,i) = noiseq;
     _EP.setZeros();
     for (int i = 0; i < 4; ++i) {
       _EP(i,i) = 1.0;
       _Estate(i,0) = 0.0; 
    }  
    break;
  case kPVSM: 
    _EH2.setZeros();
    _EH2(0,0) = _EH2(1,1) = _EH2(2,2) = _EH2(3,3) = 1.0;
    _ER.setZeros();
    for (int i = 0; i < 4; ++i)
     _ER(i,i) = noiseq;
    _EA.setZeros();
    _Eresidual.setZeros();
    _EP.setZeros();
    for (int i = 0; i < 7; ++i) {
      if (i < 4)
        _EP(i,i) = 1.0;
      else
        _EP(i,i) = INITVAR;
      _Estate(i,0) = 0.0; 
    }  
    break;
  case kPVASM:
    _EH2.setZeros();
    _EH2(0,0) = _EH2(1,1) = _EH2(2,2) = _EH2(3,3) = 1.0;
    _ER.setZeros();
    for (int i = 0; i < 4; ++i)
      _ER(i,i) = noiseq;
 
    _EA.setZeros();
    _Eresidual.setZeros();
    _EP.setZeros();
    for (int i = 0; i < 10; ++i) {
      if (i < 4)
        _EP(i,i) = 1.0;
      else
        _EP(i,i) = INITVAR;
      _Estate(i,0) = 0.0; 
    }
    break;
   default: cerr << "Problem in MMAEExtendedKalmanBase initialize" << endl;
  }
}


void 
MMAEExtendedKalmanBase::mmaeExtendedKalmanFilter(Pose *userpos, int zNum)
{
  // do the time update step using _edt and _Estate
   MMatrix epstate = _up.RK4(_Estate, _edt, _mod);

  // update the fundamental matrix, Q matrix, and the H matrix
  _EfundMat = _up.updateFundMat(epstate, _edt, _mod);
  _EQ = _up.updateQMat(epstate, _edt, _spd, _mod);
  _EH = _up.updateHMat(epstate, _mod);

  // do the a priori estimate of the covariance matrix
  MMatrix epcov = _EfundMat * (_EP * transp(_EfundMat)) + _EQ;

   // compute the Kalman Gain
   _EA = _EH * (epcov * transp(_EH)) + _ER;
   MMatrix EKG = epcov * (transp(_EH) * minverse2(_EA));
   // do the mearurment update
   MMatrix EZ(zNum,1);
   switch(_mod) {
     case kPSM:
     case kPVSM:
     case kPVASM:    
       for (int i = 0; i < 3; ++i)
	  EZ(i,0) = userpos->_orient.V()[i];
       EZ(3,0) = userpos->_orient.S();
       break;
     default: ;
   }
   _Eresidual = EZ - (_EH2 * epstate);
   _Estate = epstate + EKG * (EZ - (_EH * epstate));

   // update covariance
   _EP = epcov - (EKG * _EH * epcov);

   // better normalize the quaternion just in case
   Quat tmpq = normalize(Quat(_Estate(0,0), _Estate(1,0), _Estate(2,0), _Estate(3,0)));

   // put it back in the state vector
   for (int i = 0; i < 3; ++i)
     _Estate(i,0) = tmpq.V()[i];
   _Estate(3,0) = tmpq.S();
   
   //   cerr << " " << _Estate(0,0) << " " << sqrt(_EP(0,0)) << endl;
}
